/*
 * Copyright (c) 2018 Alibaba Group Holding Ltd. All rights reserved.
 */

'use strict';

const common = require('linkedge-common-js');
const Try = require('./try');

const envVar = common.envVars;
const IPCWrapper = common.IPCWrapper;

const MY_FUNCTION_ID = envVar.MY_FUNCTION_ID;
const AUTH_TOKEN = envVar.AUTH_TOKEN;

const ENV_VARS = {
  MY_FUNCTION_ID,
  AUTH_TOKEN,
};

function checkEnvVars(vars) {
  for (var key in vars) {
    if (!ENV_VARS[key]) {
      var err = new Error(`"${key}" environment variable is missing.`);
      throw err;
    }
  }
}

// Check whether environment variables exist, or throw error.
checkEnvVars(ENV_VARS);

console.log(`Running function ${MY_FUNCTION_ID}`);

class Runtime {
  constructor(handler, functionId, authToken) {
    this._ipc = new IPCWrapper(authToken);
    this._functionId = functionId;
    try {
      var index = handler.lastIndexOf('.');
      var moduleName = handler.slice(0, index);
      var handlerName = handler.slice(index + 1);
      this._module = require(process.cwd() + "/" + moduleName);
      this._handler = this._module[handlerName];
    } catch (err) {
      console.log(`Failed to import function handler: "${handler}"`);
      throw err;
    }
  }

  start() {
    this._getTaskAndInvoke();
  }

  _getTaskAndInvoke() {
    var doInvoke = (error, invocationId, invokerContext, payload) => {
      var invokerContextTry = this._getInvokerContext(invokerContext);
      if (invokerContextTry.error) {
        console.log(TAG, `Cannot parse the given InvokerContext: ${invokerContextTry.error}`);
        return;
      }
      this._invocation = this._newInvocation(invocationId, invokerContextTry.value, payload);
      this._invocation.invoke(this._handler, this._module);
    };
    this._ipc.getTask(this._functionId, doInvoke);
  }

  _newInvocation(invocationId, invokerContext, payload) {
    return new Invocation(this._functionId, invocationId, invokerContext, payload,
      this._postResultOrError.bind(this));
  }

  _postResultOrError(error, result) {
    if (error !== undefined && error !== null) {
      this._postHandlerError(error);
    } else {
      this._postTaskResult(result);
    }
  }

  _postTaskResult(result) {
    // Support Buffer, String, and convert other objects to JSON string.
    /*if (Object.prototype.toString.call(result) !== '[object String]'
      && !(result instanceof Buffer)) {
      var valueTry = Try.JSON.stringify(result);
      if (valueTry.error) {
        console.error(`Task result is not JSON compatible: ${valueTry.error}.`);
        throw valueTry.error;
      }
      result = valueTry.value;
    }
    result = JSON.stringify(result);*/
    var callback = (err) => {
      this._getTaskAndInvoke();
    };
    this._ipc.postTaskResult(this._functionId, this._invocation.invocationId, result,
      callback);
  }

  _postHandlerError(error) {
    var errorTry = Try.JSON.stringify(this._getErrorObject(error));
    if (errorTry.error) {
      console.error(`Failed to parse handler error: ${errorTry.error}.`);
      throw errorTry.error;
    }
    var callback = (err) => {
      this._getTaskAndInvoke();
    };
    this._ipc.postHandlerError(this._functionId, this._invocation.invocationId,
      errorTry.value, callback);
  }

  _getErrorObject(error) {
    if (error instanceof Error) {
      return {
        errorMessage: error.message,
        errorType: error.message,
        stackTrace: error.stack.split('\n').slice(1).map(function (element) {
          var prefix = '    at ';
          return element.startsWith(prefix) ? element.slice(prefix.length) : element;
        }),
      };
    } else {
      return {
        errorMessage: error.toString(),
      }
    }
  }

  _getInvokerContext(invokerContext) {
    if (invokerContext) {
      var result = Try.Base64.decode(invokerContext);
      if (result.error) {
        return result;
      }
      return Try.JSON.parse(result.value.toString());
    } else {
      return Try.TryResult.newValue({});
    }
  }

}

class Invocation {
  constructor(functionId, invocationId, invokerContext, payload, postResultOrError) {
    this.functionId = functionId;
    this.invocationId = invocationId;
    this.payload = payload;
    this.postResultOrError = postResultOrError;
    this.context = {
      requestId: invocationId,
      invokerContext: {
        custom: invokerContext.custom ? invokerContext.custom : {},
      }
    };
  }

  callback(error, result) {
    setTimeout(() => this.postResultOrError(error, result), 0);
  }

  invoke(handler, module) {
    try {
      handler.apply(module, [this.payload, this.context, this.callback.bind(this)]);
    } catch (err) {
      var error = new Error(`Failed to invoke handler: ${err}`);
      console.log(error);
      this.postResultOrError(error);
    }
  }
}

function findHandlerArg(arg) {
  return arg.startsWith('--handler=');
}

function getHandler(argv) {
  var handlerArg = argv.find(findHandlerArg);
  if (handlerArg === undefined) {
    throw new Error('--handler=<handler name> is required argument.');
  }
  var index = handlerArg.indexOf('=');
  return handlerArg.slice(index + 1);
}

process.on('SIGTERM', () => {
  console.log(`Signal: SIGTERM, stopping runtime.`);
  process.exit(0);
});

var runtime = new Runtime(getHandler(process.argv), MY_FUNCTION_ID, AUTH_TOKEN);
runtime.start();



